// SPDX: BSD-2-Clause

use futures::future::{self, Either};
use futures::stream::StreamExt;
use log::{error, info};
use netlink_sys::{AsyncSocket, SocketAddr};
use rtnetlink::new_connection;

use dynip_cloudflare::{
    utils::{self, duration_to_string},
    CloudflareClient, ExitListener, MessageHandler, MAX_ERORS_IN_ROW_DEFAULT,
};
use scopeguard::defer;
use tokio::time;

const RTNLGRP_LINK: u32 = 1;
const RTNLGRP_IPV4_IFADDR: u32 = 5;

const fn nl_mgrp(group: u32) -> u32 {
    if group > 31 {
        panic!("use netlink_sys::Socket::add_membership() for this group");
    }
    if group == 0 {
        0
    } else {
        1 << (group - 1)
    }
}

#[tokio::main]
async fn main() {
    dynip_cloudflare::init_logger();
    defer! {
        log::logger().flush();
    }
    let exit_listener = ExitListener::new();

    let config = if let Some(aux) = utils::get_config().await {
        aux
    } else {
        return;
    };

    let mut cloudflare =
        match CloudflareClient::new(config.api_key, config.zone_id, config.domains).await {
            Ok(cloudflare) => cloudflare,
            Err(e) => {
                error!("Failed to initialize cloudflare client: {:?}", &e);
                return;
            }
        };

    let (mut conn, mut _handle, mut messages) = new_connection().unwrap();
    let groups = nl_mgrp(RTNLGRP_LINK) | nl_mgrp(RTNLGRP_IPV4_IFADDR);

    let addr = SocketAddr::new(0, groups);

    if let Err(e) = conn.socket_mut().socket_mut().bind(&addr) {
        error!("Failed to bind to socket: {:?}", &e);
        return;
    }

    tokio::spawn(conn);
    info!("Listening for IPv4 address changes and interface connect/disconnect events...");

    let mut message_handler = MessageHandler::new(
        &mut cloudflare,
        config.max_errors_in_row.unwrap_or(MAX_ERORS_IN_ROW_DEFAULT),
    );

    let mut interval = config.max_duration.map(|duration| {
        let mut interval = time::interval(duration);
        interval.set_missed_tick_behavior(time::MissedTickBehavior::Delay);
        interval
    });

    while !exit_listener.should_exit() {
        let tick_future = match interval.as_mut() {
            Some(interval) => Either::Left(interval.tick()),
            None => Either::Right(future::pending::<tokio::time::Instant>()),
        };

        tokio::select! {
            _ = exit_listener.notified() => break,
            _ = tick_future => {
                if let Some(duration) = config.max_duration.as_ref() {
                    let duration_string = duration_to_string(duration);
                    let log_string = format!("{} has passed since last check, checking...", duration_string.trim());
                    message_handler.log_and_check(Some(&log_string), Option::<&&str>::None).await;
                }
            }
            message = messages.next() => {
                if let Some((message, _)) = message {
                    if let Some(interval) = interval.as_mut() {
                        interval.reset();
                    }
                    message_handler.handle_message(message).await;
                }
            }
        }
    }
}
